;;; グローバル変数/定数定義
;; √2。斜め移動時の補正に使う
(defconstant +sqrt-2+ (sqrt 2))

(load "key-state.lisp" :external-format :utf-8)

(defun load-png-image (source-file)
  (sdl:convert-to-display-format :surface (sdl:load-image source-file)
                                 :enable-alpha t
                                 :pixel-alpha t))

(defun main ()
  (sdl:with-init ()
    (sdl:window 640 480 :title-caption "てすと")
    (setf (sdl:frame-rate) 60)
    (let ((current-key-state (make-instance 'key-state))
          (x 0) ; 自機X座標
          (y 0) ; 自機Y座標
          (speed 5) ; 速さ。1フレームあたりの移動量
          ;; 自機の画像をロード
          (player-img (load-png-image "player.png")))

      (sdl:update-display)

      (sdl:with-events ()
        (:quit-event () t)
        (:key-down-event (:key key)
          (if (sdl:key= key :sdl-key-escape)
            (sdl:push-quit-event)
            (update-key-state key t current-key-state)))
        (:key-up-event (:key key)
          (update-key-state key nil current-key-state))
        (:idle ()
          (sdl:clear-display sdl:*black*)
          ;; 移動キーの操作を移動量に反映
          (let ((dx 0) (dy 0)) ; X方向とY方向の移動量
            (with-slots (up down right left) current-key-state
              (cond (right  (incf dx speed))
                    (left   (decf dx speed)))
              (cond (up   (decf dy speed))
                    (down (incf dy speed))))
            ;; 斜め移動の場合の処理。移動量を補正します
            (when (and (/= dx 0) (/= dy 0))
              (setf dx (/ dx +sqrt-2+) dy (/ dy +sqrt-2+)))
            ;; 移動量を自機の座標に反映
            (incf x dx)
            (incf y dy))
          ;; 自機を描画
          (sdl:draw-surface-at-* player-img (round x) (round y))
          (sdl:update-display))))))

