;; 総称関数
(defgeneric add-enemy (enemy-manager enemy)
  (:documentation "管理する敵を追加する"))
(defgeneric update-enemies (enemy-manager)
  (:documentation "敵の状態を更新する"))
(defgeneric draw-enemies (enemy-manager game-field)
  (:documentation "敵を描画する"))
(defgeneric gc-enemies (enemy-manager game-field)
  (:documentation "不要な敵を削除する"))
(defgeneric hit-test-enemies (enemy-manager region &key)
  (:documentation "敵との当たり判定を行う。"))

;; 敵管理クラス
(defclass enemy-manager ()
  ((enemies
    :initform nil
    :documentation "管理している敵たち")))

(defmethod add-enemy ((enemy-manager enemy-manager) (enemy enemy))
  (with-slots (enemies) enemy-manager
    (push enemy enemies)))

(defmethod update-enemies ((enemy-manager enemy-manager))
  (with-slots (enemies) enemy-manager
    (dolist (enemy enemies)
      (update-enemy enemy))))

(defmethod draw-enemies ((enemy-manager enemy-manager) (game-field game-field))
  (with-slots (enemies) enemy-manager
    (dolist (enemy enemies)
      (draw-enemy enemy game-field))))

(defmethod gc-enemies ((enemy-manager enemy-manager) (game-field game-field))
  (with-slots (enemies) enemy-manager
    (setf enemies
          (delete-if #'(lambda (enemy)
                         (out-of-field-p enemy game-field))
                     enemies))))

(defmethod hit-test-enemies ((enemy-manager enemy-manager)
                             (region circle-region)
                             &key 
                             (region-type 'attack)
                             (on-hit-func
                               #'(lambda (enemy)
                                   (declare (ignore enemy))
                                   nil)))
  (let ((region-getter (cond ((eq region-type 'attack) #'attack-region)
                             ((eq region-type 'damage) #'damage-region))))
    (with-slots (enemies) enemy-manager
      (some #'(lambda (enemy)
                (if (hit-test (funcall region-getter enemy) region)
                  (progn
                    (funcall on-hit-func enemy)
                    t)
                  nil))
            enemies))))
