;; 総称関数
(defgeneric update-enemy (enemy)
  (:documentation "敵の状態を更新する"))
(defgeneric draw-enemy (enemy game-field)
  (:documentation "敵を描画する"))
(defgeneric out-of-field-p (enemy game-field)
  (:documentation "敵が画面外にいるかどうか判定"))

;; 敵クラス
(defclass enemy ()
  ((x
    :initarg :x
    :initform 0
    :documentation "中心X座標")
   (y
    :initarg :y
    :initform 0
    :documentation "中心Y座標")
   (vx
    :initarg :vx
    :initform 0
    :documentation "X方向速度")
   (vy
    :initarg :vy
    :initform 0
    :documentation "Y方向速度")
   (width
    :initarg :width
    :initform 128
    :documentation "幅")
   (height
    :initarg :height
    :initform 128
    :documentation "高さ")
   (attack-region-radius
    :initarg :attack-radius
    :initform 10
    :documentation "攻撃判定半径")
   (damage-region-radius
    :initarg :damage-radius
    :initform 10
    :documentation "ダメージ判定半径")
   (attack-region
    :accessor attack-region
    :documentation "攻撃判定")
   (damage-region
    :accessor damage-region
    :documentation "ダメージ判定")
   (image
    :initarg :image
    :initform (error "敵画像ファイルを指定しろ")
    :documentation "画像")))

;; 敵クラス初期化処理
(defmethod initialize-instance :after ((enemy enemy) &key)
  (with-slots (attack-region attack-region-radius x y) enemy
    (setf attack-region (make-instance 'circle-region
                                       :x x
                                       :y y
                                       :r attack-region-radius)))
  (with-slots (damage-region damage-region-radius x y) enemy
    (setf damage-region (make-instance 'circle-region
                                       :x x
                                       :y y
                                       :r damage-region-radius))))

(defmethod update-enemy ((enemy enemy))
  (flet ((update-hit-region (x y region)
           (setf (circle-x region) x)
           (setf (circle-y region) y)))
    (with-slots (x y vx vy) enemy
      (incf x vx)
      (incf y vy)
      ;; 当たり判定の座標更新
      (update-hit-region x y (attack-region enemy))
      (update-hit-region x y (damage-region enemy)))))

(defmethod draw-enemy ((enemy enemy) (game-field game-field))
  (flet ((draw-hit-region (region color)
           (sdl:draw-circle-* (round (circle-x region) )
                              (round (circle-y region) )
                              (round (circle-radius region))
                              :color color)))
    (with-slots (x y width height image) enemy
      (sdl:draw-surface-at-* image
                             (round (+ (game-field-x game-field)
                                       (- x (/ width 2))))
                             (round (+ (game-field-y game-field)
                                       (- y (/ height 2))))))
    ;; デバッグ用。当たり判定描画
    (draw-hit-region (damage-region enemy) sdl:*green*)
    (draw-hit-region (attack-region enemy) sdl:*red*)))

(defmethod out-of-field-p ((enemy enemy) (game-field game-field))
  (with-slots (x y width height) enemy
    (let ((width/2 (/ width 2))
          (height/2 (/ height 2)))
      (or (< (+ x width/2) 0)
          (< (+ y height/2) 0)
          (> (- x width/2) (game-field-width game-field))
          (> (- y height/2) (game-field-height game-field))))))
