(defgeneric add-bullet (bullet-manager bullet)
  (:documentation "弾を追加する"))
(defgeneric update-bullets (bullet-manager)
  (:documentation "管理している弾を更新する"))
(defgeneric draw-bullets (bullet-manager game-field)
  (:documentation "弾を描画する"))
(defgeneric gc-bullets (bullet-manager game-field)
  (:documentation "不要になった弾を削除する"))
(defgeneric hit-test-bullets (bullet-manager region &key)
  (:documentation "管理している弾との当たり判定"))

(defmacro do-bullets ((bullet-var bullet-manager) &body body)
  (let ((bullets-symbol (gensym)))
  `(with-slots ((,bullets-symbol bullets)) ,bullet-manager
     (dolist (,bullet-var ,bullets-symbol)
       ,@body))))

(defclass bullet-manager ()
  ((bullets
    :initform nil
    :documentation "管理している弾")))

(defmethod add-bullet ((bullet-manager bullet-manager) (bullet bullet))
  (with-slots (bullets) bullet-manager
    (push bullet bullets)))

(defmethod update-bullets ((bullet-manager bullet-manager))
  (do-bullets (bullet bullet-manager)
    (update-bullet bullet)))

(defmethod draw-bullets ((bullet-manager bullet-manager)
                         (game-field game-field))
  (do-bullets (bullet bullet-manager)
    (draw-bullet bullet game-field)))

(defmethod gc-bullets ((bullet-manager bullet-manager)
                       (game-field game-field))
  (with-slots (bullets) bullet-manager
    (setf bullets
          (delete-if #'(lambda (bullet)
                         (or (not (alive bullet))
                             (out-of-field-p bullet game-field)))
                     bullets))))

(defmethod hit-test-bullets ((bullet-manager bullet-manager)
                             (region circle-region)
                             &key 
                             (on-hit-func
                               #'(lambda (bullet)
                                   (declare (ignore bullet))
                                   nil)))
  (with-slots (bullets) bullet-manager
    (some #'(lambda (bullet)
              (if (hit-test (attack-region bullet) region)
                (progn
                  (funcall on-hit-func bullet)
                  t)
                nil))
          bullets)))
