;;; グローバル変数/定数定義
;; √2。斜め移動時の補正に使う
(defconstant +sqrt-2+ (sqrt 2))

(load "key-state.lisp" :external-format :utf-8)
(load "game-field.lisp" :external-format :utf-8)
(load "hit-test.lisp" :external-format :utf-8)
(load "bullet.lisp" :external-format :utf-8)
(load "bullet-manager.lisp" :external-format :utf-8)
(load "player.lisp" :external-format :utf-8)
(load "enemy.lisp" :external-format :utf-8)
(load "enemy-manager.lisp" :external-format :utf-8)

(defun load-png-image (source-file)
  (sdl:convert-to-display-format :surface (sdl:load-image source-file)
                                 :enable-alpha t
                                 :pixel-alpha t))

(defun main ()
  (sdl:with-init ()
    (sdl:window 640 480 :title-caption "てすと")
    (sdl:initialise-default-font sdl:*font-10x20*) ; フォント初期化
    (setf (sdl:frame-rate) 60)
    (let* ((current-key-state (make-instance 'key-state))
           ;; ゲーム画面オブジェクトを生成
           (game-field (make-instance 'game-field
                                      :width 640
                                      :height 480
                                      :x 0
                                      :y 0))
           ;; 自機オブジェクトを生成
           (player (make-instance 'player
                                  :x (/ (game-field-width game-field) 2)
                                  :y (/ (game-field-height game-field) 2)
                                  :width 64
                                  :height 64
                                  :speed 5
                                  :damage-radius 10
                                  :image (load-png-image "player.png")
                                  :bullet-image
                                  (load-png-image "bullet-2.png")))
           ;; 敵管理クラスを生成
           (enemy-manager (make-instance 'enemy-manager))
           ;;;; 自機弾管理
           (player-bullet-manager (make-instance 'bullet-manager))
           ;;;; 敵弾管理
           (enemy-bullet-manager (make-instance 'bullet-manager))
           ;; 自機存命フラグ
           (player-is-alive t))
      ;;敵を生成
      (dotimes (i 10)
        (add-enemy enemy-manager
                   (make-instance 
                     'enemy
                     :x (+ 32 (* i (/ (game-field-width game-field) 10)))
                     :y 0
                     :width 64
                     :height 64
                     :vx 1
                     :vy 1
                     :attack-radius 30
                     :damage-radius 20
                     :image (load-png-image "enemy.png")
                     :bullet-image (load-png-image "bullet-1.png"))))
      (sdl:update-display)

      (sdl:with-events ()
        (:quit-event () t)
        (:key-down-event (:key key)
          (if (sdl:key= key :sdl-key-escape)
            (sdl:push-quit-event)
            (update-key-state key t current-key-state)))
        (:key-up-event (:key key)
          (update-key-state key nil current-key-state))
        (:idle ()
          (sdl:clear-display sdl:*black*)
          ;; 自機が生きている時だけゲーム状態を更新
          (when player-is-alive
            ;; 移動キーの操作を移動量に反映
            (update-player player current-key-state
                           game-field player-bullet-manager)
            ;; 敵の状態を更新
            (update-enemies enemy-manager enemy-bullet-manager)
            ;; 自機弾更新
            (update-bullets player-bullet-manager)
            ;; 敵弾更新
            (update-bullets enemy-bullet-manager)
            ;; 自機弾対敵当たり判定
            (do-bullets (bullet player-bullet-manager)
              (when (hit-test-enemies enemy-manager
                                      (attack-region bullet)
                                      :region-type 'damage
                                      :on-hit-func
                                      #'(lambda (enemy)
                                          ;; 当たった敵は死ぬ
                                          (setf (alive enemy) nil)))
                ;; 敵を当てた弾を消す
                (setf (alive bullet) nil)))
            ;; 敵弾対自機当たり判定
            (when (hit-test-bullets enemy-bullet-manager
                                    (damage-region player))
              (setf player-is-alive nil))
            ;; 画面外に出た敵や死んだ敵を成仏
            (gc-enemies enemy-manager game-field)
            (gc-bullets player-bullet-manager game-field)
            (gc-bullets enemy-bullet-manager game-field)
            ;; 敵対自機の当たり判定
            (when (hit-test-enemies enemy-manager
                                    (damage-region player)
                                    :region-type 'attack)
              (setf player-is-alive nil)))

          ;; 自機弾を描画
          (draw-bullets player-bullet-manager game-field)
          ;; 自機を描画
          (draw-player player game-field)
          ;; 敵を描画
          (draw-enemies enemy-manager game-field)
          ;; 敵弾を描画
          (draw-bullets enemy-bullet-manager game-field)
          ;; 死んだら残念な表示
          (when (not player-is-alive)
            (sdl:draw-string-solid-* "You died! Push Esc key..."
                                     220 240))
          (sdl:update-display))))))
