;; 総称関数
(defgeneric update-bullet (bullet)
  (:documentation "弾の状態を更新する"))
(defgeneric draw-bullet (bullet game-field)
  (:documentation "弾を描画する"))
(defgeneric out-of-field-p (bullet game-field)
  (:documentation "弾が画面外にいるかどうか判定"))

;; 弾クラス
(defclass bullet ()
  ((x
    :initarg :x
    :initform 0
    :documentation "中心X座標")
   (y
    :initarg :y
    :initform 0
    :documentation "中心X座標")
   (vx
    :initarg :vx
    :initform 0
    :documentation "X方向速度")
   (vy
    :initarg :vy
    :initform 0
    :documentation "Y方向速度")
   (width
    :initarg :width
    :initform 16
    :documentation "幅")
   (height
    :initarg :height
    :initform 16
    :documentation "高さ")
   (attack-region-radius
    :initarg :attack-radius
    :initform 10
    :documentation "攻撃判定半径")
   (attack-region
    :accessor attack-region
    :documentation "攻撃判定")
   (alive
    :accessor alive
    :initform t
    :documentation "弾の存命フラグ")
   (image
    :initarg :image
    :initform (error "弾画像ファイルを指定しろ")
    :documentation "画像")))

;; 弾クラス初期化処理
(defmethod initialize-instance :after ((bullet bullet) &key)
  (with-slots (attack-region attack-region-radius x y) bullet
    (setf attack-region (make-instance 'circle-region
                                       :x x
                                       :y y
                                       :r attack-region-radius))))

(defmethod update-bullet ((bullet bullet))
  (when (alive bullet)
    (flet ((update-hit-region (x y region)
             (setf (circle-x region) x)
             (setf (circle-y region) y)))
      (with-slots (x y vx vy) bullet
        (incf x vx)
        (incf y vy)
        (update-hit-region x y (attack-region bullet))))))

(defmethod draw-bullet ((bullet bullet) (game-field game-field))
  (with-slots (x y width height image) bullet
    (sdl:draw-surface-at-* image
                           (round (+ (game-field-x game-field)
                                     (- x (/ width 2))))
                           (round (+ (game-field-y game-field)
                                     (- y (/ height 2)))))))

(defmethod out-of-field-p ((bullet bullet) (game-field game-field))
  (with-slots (x y width height) bullet
    (let ((width/2 (/ width 2))
          (height/2 (/ height 2)))
      (or (< (+ x width/2) 0)
          (< (+ y height/2) 0)
          (> (- x width/2) (game-field-width game-field))
          (> (- y height/2) (game-field-height game-field))))))
