;;; グローバル変数/定数定義
;; √2。斜め移動時の補正に使う
(defconstant +sqrt-2+ (sqrt 2))

(load "loader.lisp" :external-format :utf-8)
(load "key-state.lisp" :external-format :utf-8)
(load "game-field.lisp" :external-format :utf-8)
(load "hit-test.lisp" :external-format :utf-8)
(load "bullet.lisp" :external-format :utf-8)
(load "bullet-manager.lisp" :external-format :utf-8)
(load "player.lisp" :external-format :utf-8)
(load "enemy.lisp" :external-format :utf-8)
(load "enemy-manager.lisp" :external-format :utf-8)
(load "stage.lisp" :external-format :utf-8)

(defun main ()
  (sdl:with-init ()
    (sdl:window 640 480 :title-caption "てすと")
    (sdl:initialise-default-font sdl:*font-10x20*)
    (setf (sdl:frame-rate) 60)
    ;; ローダー生成
    (let ((loader (make-instance 'image-loader)))
      ;; 画像を読み込んでいく
      (add-image loader :player "player.png")
      (add-image loader :enemy "enemy.png")
      (add-image loader :player-bullet "bullet-2.png")
      (add-image loader :enemy-bullet "bullet-1.png")
      (let* ((current-key-state (make-instance 'key-state))
             (game-field (make-instance 'game-field
                                        :width 640
                                        :height 480
                                        :x 0
                                        :y 0))
             (player (make-instance 'player
                                    :x (/ (game-field-width game-field) 2)
                                    :y (/ (game-field-height game-field) 2)
                                    :width 64
                                    :height 64
                                    :speed 5
                                    :damage-radius 10
                                    :image (get-image loader :player)
                                    :bullet-image
                                    (get-image loader :player-bullet)))
             (enemy-manager (make-instance 'enemy-manager))
             (player-bullet-manager (make-instance 'bullet-manager))
             (enemy-bullet-manager (make-instance 'bullet-manager))
             ;; ステージ
             (stage (make-instance 'stage
                                   :schedule 
                                   '((:frame 0   :x 0 :y 0 :vx 2 :vy 2)
                                     (:frame 60  :x 0 :y 0 :vx 1.5 :vy 2)
                                     (:frame 120 :x 0 :y 0 :vx 1 :vy 2)
                                     (:frame 180 :x 0 :y 0 :vx 0.5 :vy 2))))
             (player-is-alive t))
        (sdl:update-display)

        (sdl:with-events ()
          (:quit-event () t)
          (:key-down-event (:key key)
            (if (sdl:key= key :sdl-key-escape)
              (sdl:push-quit-event)
              (update-key-state key t current-key-state)))
          (:key-up-event (:key key)
            (update-key-state key nil current-key-state))
          (:idle ()
            (sdl:clear-display sdl:*black*)
            (when player-is-alive
              ;; ステージ更新
              (update-stage stage enemy-manager loader)
              (update-player player current-key-state
                             game-field player-bullet-manager)
              (update-enemies enemy-manager enemy-bullet-manager)
              (update-bullets player-bullet-manager)
              (update-bullets enemy-bullet-manager)
              (do-bullets (bullet player-bullet-manager)
                (when (hit-test-enemies enemy-manager
                                        (attack-region bullet)
                                        :region-type 'damage
                                        :on-hit-func
                                        #'(lambda (enemy)
                                            (setf (alive enemy) nil)))
                  (setf (alive bullet) nil)))
              (when (hit-test-bullets enemy-bullet-manager
                                      (damage-region player))
                (setf player-is-alive nil))
              (gc-enemies enemy-manager game-field)
              (gc-bullets player-bullet-manager game-field)
              (gc-bullets enemy-bullet-manager game-field)
              (when (hit-test-enemies enemy-manager
                                      (damage-region player)
                                      :region-type 'attack)
                (setf player-is-alive nil)))
          
            (draw-bullets player-bullet-manager game-field)
            (draw-player player game-field)
            (draw-enemies enemy-manager game-field)
            (draw-bullets enemy-bullet-manager game-field)
            (when (not player-is-alive)
              (sdl:draw-string-solid-* "You died! Push Esc key..."
                                       220 240))
            (sdl:update-display)))))))
