;; 総称関数
(defgeneric update-player (player key-state game-field bullet-manager)
  (:documentation "自機の状態を更新する"))
(defgeneric draw-player (player game-field)
  (:documentation "自機を描画する"))

;; 自機クラス
(defclass player ()
  ((x
    :initarg :x
    :initform 0
    :documentation "中心X座標")
   (y
    :initarg :y
    :initform 0
    :documentation "中心Y座標")
   (width
    :initarg :width
    :initform 64
    :documentation "幅")
   (height
    :initarg :height
    :initform 64
    :documentation "高さ")
   (speed
    :initarg :speed
    :initform 5
    :documentation "速さ")
   (damage-region-radius
    :initarg :damage-radius
    :initform 10
    :documentation "ダメージ判定半径")
   (damage-region
    :accessor damage-region
    :documentation "ダメージ判定")
   (update-frame
    :accessor update-frame
    :initform 0
    :documentation "自機更新の時間軸")
   (bullet-image
    :initarg :bullet-image
    :initform (error "自機弾画像ファイルを指定しろ")
    :documentation "弾の画像")
   (image
    :initarg :image
    ;; 初期化するときに必ず画像ファイルを指定してください＞＜
    :initform (error "自機画像ファイルを指定しろ")
    :documentation "画像")))

;; 自機クラス初期化処理
(defmethod initialize-instance :after ((player player) &key)
  (with-slots (damage-region damage-region-radius x y) player
    (setf damage-region (make-instance 'circle-region
                                       :x x
                                       :y y
                                       :r damage-region-radius))))

(defmethod update-player ((player player)
                          (key-state key-state)
                          (game-field game-field)
                          (bullet-manager bullet-manager))
  ;; 弾発射ロジック
  (with-slots (shoot) key-state
    (when (and shoot (zerop (mod (update-frame player) 4)))
      (with-slots (x y bullet-image) player
        (add-bullet bullet-manager 
                    (make-instance 'bullet
                                   :x x
                                   :y y
                                   :vx 0
                                   :vy -10
                                   :width 16
                                   :height 16
                                   :attack-radius 8
                                   :image bullet-image)))))
  (let ((dx 0) (dy 0)) ; X方向とY方向の移動量
    (with-slots (up down right left) key-state
      (with-slots (speed) player
        (cond (right  (incf dx speed))
              (left   (decf dx speed)))
        (cond (up   (decf dy speed))
              (down (incf dy speed)))))
    ;; 斜め移動の場合の処理。移動量を補正します
    (when (and (/= dx 0) (/= dy 0))
      (setf dx (/ dx +sqrt-2+) dy (/ dy +sqrt-2+)))
    ;; 移動量を自機の座標に反映
    (with-slots (x y width height) player
      (incf x dx)
      (incf y dy)
      ;; ゲーム画面から出るのを防ぐ
      (let ((field-width (game-field-width game-field))
            (field-height (game-field-height game-field))
            (width/2 (/ width 2))
            (height/2 (/ height 2))
            (region (damage-region player)))
        (when (< (- x width/2) 0) (setf x width/2))
        (when (< (- y height/2) 0) (setf y height/2))
        (when (> (+ x width/2) field-width)
          (setf x (- field-width width/2)))
        (when (> (+ y height/2) field-height)
          (setf y (- field-height height/2)))
        ;; 当たり判定座標更新
        (setf (circle-x region) x)
        (setf (circle-y region) y))))
  (incf (update-frame player)))

(defmethod draw-player ((player player) (game-field game-field))
  (with-slots (x y width height image) player
    (sdl:draw-surface-at-* image
                           (round (+ (game-field-x game-field)
                                     (- x (/ width 2))))
                           (round (+ (game-field-y game-field)
                                     (- y (/ height 2)))))))
