;; 総称関数
(defgeneric update-stage (stage enemy-manager loader)
  (:documentation "ステージの更新"))

;; ステージクラス定義
(defclass stage ()
  ((frame
    :initform 0
    :accessor frame
    :documentation "ステージの時間軸")
   (current-schedule-index
    :initform 0
    :documentation "現在のスケジュールを指すインデックス")
   (schedule
    :initarg :schedule
    :initform nil
    :documentation "敵が出現するスケジュール")))

(defun get-frame-from-schedule (s)
  (getf s :frame))

(defun create-enemy-from-schedule (s loader)
  (make-instance 
    'enemy
    :x (getf s :x)
    :y (getf s :y)
    :width 64
    :height 64
    :vx (getf s :vx)
    :vy (getf s :vy)
    :attack-radius 30
    :damage-radius 20
    :image (get-image loader :enemy)
    :bullet-image (get-image loader :enemy-bullet)))

(defmethod initialize-instance :after ((stage stage) &key)
  (with-slots (schedule) stage
    (setf schedule (sort schedule #'< :key #'get-frame-from-schedule))))

(defmethod update-stage ((stage stage)
                         (enemy-manager enemy-manager)
                         (loader image-loader))
  (with-slots (schedule frame current-schedule-index) stage
    (loop for s in (nthcdr current-schedule-index schedule)
          if (= (get-frame-from-schedule s) frame)
          do (progn
               (incf current-schedule-index)
               (add-enemy enemy-manager
                          (create-enemy-from-schedule s loader))))
    (incf frame)))
